#include "drv_gyro.h"

#include "hal_twi.h"
#include "io_cfg.h"
#include "nordic_common.h"
#include "nrf_assert.h"
#include "nrf_delay.h"
#include "nrf_error.h"

#define MPU6050_ADDR              0x68
#define ADDRESS_WHO_AM_I          0x75 //!< WHO_AM_I register identifies the device. Expected value is 0x68.
#define ADDRESS_SIGNAL_PATH_RESET 0x68 //!< 

//-----------------------------------------------------------------------------
// Register map
//-----------------------------------------------------------------------------
#define MPU_CONFIG        0x1A
#define GYRO_CONFIG       0x1B
#define ACCEL_CONFIG      0x1C
#define ZRMOT_THR         0x21
#define ZRMOT_DUR         0x22
#define ACCEL_XOUT_H      0x3B
#define INT_PIN_CFG       0x37
#define INT_ENABLE        0x38
#define INT_STATUS        0x3A
#define MOT_DETECT_STATUS 0x61
#define MOT_DETECT_CTRL   0x69
#define PWR_MGMT_1        0x6B
#define PWR_MGMT_2        0x6C
#define MPU_WHO_AM_I      0x75

static const uint8_t expected_who_am_i = 0x68U; //!< Expected value to get from WHO_AM_I register.
static uint8_t       s_device_address; //!< Device address in bits [7:1]
const static uint8_t s_register_config[][2] =  {
                                                   {PWR_MGMT_1,0x00},
                                                   {MPU_CONFIG,0x01},
                                                   {GYRO_CONFIG,0x18},
                                                   {ACCEL_CONFIG,0x02},
                                                   {ZRMOT_THR,0x07},
                                                   {ZRMOT_DUR,0x4E},
                                               };

static __inline bool mpu6050_register_write(uint8_t register_address, uint8_t value)
{	
    uint8_t w2_data[2];

    w2_data[0] = register_address;
    w2_data[1] = value;
    return hal_twi_transfer(s_device_address, w2_data, 2, TWI_ISSUE_STOP);	  
}

static __inline bool mpu6050_register_read(uint8_t register_address, uint8_t *destination, uint8_t number_of_bytes)
{
    bool    transfer_succeeded;
    uint8_t address = register_address;
    
    transfer_succeeded = hal_twi_transfer(s_device_address, &address, 1, TWI_DONT_ISSUE_STOP);
    transfer_succeeded &= hal_twi_transfer(s_device_address|TWI_READ_BIT, destination, number_of_bytes, TWI_ISSUE_STOP);
    return transfer_succeeded;
}

static bool mpu6050_verify_product_id(void)
{
  uint8_t who_am_i;
  
  if (mpu6050_register_read(ADDRESS_WHO_AM_I, &who_am_i, 1))
  {
    if (who_am_i != expected_who_am_i)
    {
      return false;
    }
    else
    {
      return true;
    }
  }
  else
  {
    return false;
  }
}

uint32_t drv_gyro_init(void)
{
    hal_twi_init_t twi_params;
    uint32_t       err_code;
    uint8_t        reset_value;
    bool           transfer_succeeded = true;
    int            i;
    
    s_device_address = (uint8_t)(MPU6050_ADDR << 1);
    
    // 2-Wire initialization
    twi_params.scl_pin   = IO_TWI_SCL;
    twi_params.sda_pin   = IO_TWI_SDA;
    twi_params.frequency = hal_twi_freq_400khz;
    twi_params.ppi_chn   = 3;
    
    err_code = hal_twi_init(&twi_params);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }
    
    // Read and verify product ID
    transfer_succeeded &= mpu6050_verify_product_id();
    if (!transfer_succeeded)
    {
        return NRF_ERROR_NOT_FOUND;
    }

    // Do a reset on signal paths
    reset_value = 0x04U | 0x02U | 0x01U; // Resets gyro, accelerometer and temperature sensor signal paths
    transfer_succeeded &= mpu6050_register_write(ADDRESS_SIGNAL_PATH_RESET, reset_value);

    // Setting device configuration
    for (i = 0; i < (sizeof(s_register_config) / sizeof(s_register_config[0])); ++i)
    {
        uint8_t read;
        
        transfer_succeeded &= mpu6050_register_write(s_register_config[i][0], s_register_config[i][1]);
        transfer_succeeded &= mpu6050_register_read(s_register_config[i][0], &read, sizeof(read));
        
        transfer_succeeded &= (s_register_config[i][1] == read); 
    }
    
    if (!transfer_succeeded)
    {
        return NRF_ERROR_NOT_FOUND;
    }
    
    
    // Gyro powered down by default to save power when not used
    return drv_gyro_enable(false);
}

uint32_t drv_gyro_enable(bool p_enable)
{
    uint8_t reg_val;
    bool success;
    
    reg_val = p_enable ? 0x00 : BIT_6;
    success = mpu6050_register_write(PWR_MGMT_1, reg_val);
    
    return (success ? NRF_SUCCESS : NRF_ERROR_INTERNAL);
}

uint32_t drv_gyro_read(t_struct_AIR_MOTION_ProcessDeltaSamples* p_samples, bool* p_mot_detect, bool p_format_workaround)
{
    bool success = true;
    uint8_t buf[14];
    uint8_t mot_detect_status;
    
    success &= mpu6050_register_read(ACCEL_XOUT_H, buf, 14);
    success &= mpu6050_register_read(MOT_DETECT_STATUS, &mot_detect_status, 1);
    
    if (!success)
    {
        return NRF_ERROR_INTERNAL;
    }
    
    if (p_format_workaround)    
    {
        // 12-bit signed gyro and accelerometer values
        p_samples->GyroSamples.X = ((int16_t)  ((buf[10] << 8) | buf[11])) / 16;
        p_samples->GyroSamples.Y = -((int16_t) ((buf[8]  << 8) | buf[9]))  / 16;
        p_samples->GyroSamples.Z = ((int16_t)  ((buf[12] << 8) | buf[13])) / 16;
        p_samples->AccSamples.X  = ((int16_t)  ((buf[2]  << 8) | buf[3]))  / 16;
        p_samples->AccSamples.Y  = -((int16_t) ((buf[0]  << 8) | buf[1]))  / 16;
        p_samples->AccSamples.Z  = ((int16_t)  ((buf[4]  << 8) | buf[5]))  / 16;
    }
    else
    {
        // 12-bit signed gyro and accelerometer values
        p_samples->GyroSamples.X = ((int16_t) ((buf[8]  << 8) | buf[9]))  / 16;
        p_samples->GyroSamples.Y = ((int16_t) ((buf[10] << 8) | buf[11])) / 16;
        p_samples->GyroSamples.Z = ((int16_t) ((buf[12] << 8) | buf[13])) / 16;
        p_samples->AccSamples.X  = ((int16_t) ((buf[0]  << 8) | buf[1]))  / 16;
        p_samples->AccSamples.Y  = ((int16_t) ((buf[2]  << 8) | buf[3]))  / 16;
        p_samples->AccSamples.Z  = ((int16_t) ((buf[4]  << 8) | buf[5]))  / 16;
    }

    if (mot_detect_status & BIT_0)
    {
        *p_mot_detect = false;
    }
    else
    {
        *p_mot_detect = true;
    }
  
    return NRF_SUCCESS;
}
